@@ -4,7 +4,7 @@
 #include <petscksp.h> /*I "petscksp.h" I*/
 #include <../src/mat/impls/aij/mpi/mpiaij.h>
 #include <../src/mat/impls/aij/seq/kokkos/aijkok.hpp>
-#include "petscsection.h"
+#include <petscsection.h>
 #include <petscdmcomposite.h>
 
 #include <../src/mat/impls/aij/seq/aij.h>
@@ -49,8 +49,7 @@ KOKKOS_INLINE_FUNCTION PetscErrorCode MatMult(const team_member team, const Pets
     const PetscInt    *aj   = glb_Aaj + glb_Aai[rowa]; // global
     const PetscScalar *aa   = glb_Aaa + glb_Aai[rowa];
     PetscScalar        sum;
-    Kokkos::parallel_reduce(
-      Kokkos::ThreadVectorRange(team, n), [=](const int i, PetscScalar &lsum) { lsum += aa[i] * x_loc[r[aj[i]] - start]; }, sum);
+    Kokkos::parallel_reduce(Kokkos::ThreadVectorRange(team, n), [=](const int i, PetscScalar &lsum) { lsum += aa[i] * x_loc[r[aj[i]] - start]; }, sum);
     Kokkos::single(Kokkos::PerThread(team), [=]() { y_loc[rowb - start] = sum; });
   });
   team.team_barrier();
@@ -160,13 +159,12 @@ static KOKKOS_INLINE_FUNCTION PetscErrorCode BJSolve_TFQMR(const team_member tea
     XX[rowb - start] = 0;
   });
   team.team_barrier();
-  parallel_reduce(
-    Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &lsum) { lsum += R[idx] * PetscConj(R[idx]); }, dpi);
+  parallel_reduce(Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &lsum) { lsum += R[idx] * PetscConj(R[idx]); }, dpi);
   team.team_barrier();
   r0 = dp = PetscSqrtReal(PetscRealPart(dpi));
   // diagnostics
 #if defined(PETSC_USE_DEBUG) && !defined(PETSC_HAVE_SYCL)
-  if (monitor) Kokkos::single(Kokkos::PerTeam(team), [=]() { printf("%3d KSP Residual norm %14.12e \n", 0, (double)dp); });
+  if (monitor) Kokkos::single(Kokkos::PerTeam(team), [=]() { printf("%3d KSP Residual norm %14.12e\n", 0, (double)dp); });
 #endif
   if (dp < atol) {
     metad->reason = KSP_CONVERGED_ATOL_NORMAL;
@@ -189,8 +187,7 @@ static KOKKOS_INLINE_FUNCTION PetscErrorCode BJSolve_TFQMR(const team_member tea
   dpold  = dp;
 
   /* rhoold = (r,rp)     */
-  parallel_reduce(
-    Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &dot) { dot += R[idx] * PetscConj(RP[idx]); }, rhoold);
+  parallel_reduce(Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &dot) { dot += R[idx] * PetscConj(RP[idx]); }, rhoold);
   team.team_barrier();
   parallel_for(Kokkos::TeamVectorRange(team, Nblk), [=](int idx) {
     U[idx] = R[idx];
@@ -204,8 +201,7 @@ static KOKKOS_INLINE_FUNCTION PetscErrorCode BJSolve_TFQMR(const team_member tea
   it = 0;
   do {
     /* s <- (v,rp)          */
-    parallel_reduce(
-      Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &dot) { dot += V[idx] * PetscConj(RP[idx]); }, s);
+    parallel_reduce(Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &dot) { dot += V[idx] * PetscConj(RP[idx]); }, s);
     team.team_barrier();
     if (s == 0) {
       metad->reason = KSP_CONVERGED_HAPPY_BREAKDOWN;
@@ -226,8 +222,7 @@ static KOKKOS_INLINE_FUNCTION PetscErrorCode BJSolve_TFQMR(const team_member tea
     /* r <- r - a K (u + q) */
     parallel_for(Kokkos::TeamVectorRange(team, Nblk), [=](int idx) { R[idx] = R[idx] - a * AUQ[idx]; });
     team.team_barrier();
-    parallel_reduce(
-      Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &lsum) { lsum += R[idx] * PetscConj(R[idx]); }, dpi);
+    parallel_reduce(Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &lsum) { lsum += R[idx] * PetscConj(R[idx]); }, dpi);
     team.team_barrier();
     dp = PetscSqrtReal(PetscRealPart(dpi));
     for (m = 0; m < 2; m++) {
@@ -250,7 +245,7 @@ static KOKKOS_INLINE_FUNCTION PetscErrorCode BJSolve_TFQMR(const team_member tea
       team.team_barrier();
       dpest = PetscSqrtReal(2 * it + m + 2.0) * tau;
 #if defined(PETSC_USE_DEBUG) && !defined(PETSC_HAVE_SYCL)
-      if (monitor && m == 1) Kokkos::single(Kokkos::PerTeam(team), [=]() { printf("%3d KSP Residual norm %14.12e \n", it + 1, (double)dpest); });
+      if (monitor && m == 1) Kokkos::single(Kokkos::PerTeam(team), [=]() { printf("%3d KSP Residual norm %14.12e\n", it + 1, (double)dpest); });
 #endif
       if (dpest < atol) {
         metad->reason = KSP_CONVERGED_ATOL_NORMAL;
@@ -284,8 +279,7 @@ static KOKKOS_INLINE_FUNCTION PetscErrorCode BJSolve_TFQMR(const team_member tea
     }
 
     /* rho <- (r,rp)       */
-    parallel_reduce(
-      Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &dot) { dot += R[idx] * PetscConj(RP[idx]); }, rho);
+    parallel_reduce(Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &dot) { dot += R[idx] * PetscConj(RP[idx]); }, rho);
     team.team_barrier();
     if (rho == 0) {
       metad->reason = KSP_CONVERGED_HAPPY_BREAKDOWN;
@@ -322,8 +316,7 @@ done:
   metad->its = it;
   if (1) {
     int nnz;
-    parallel_reduce(
-      Kokkos::TeamVectorRange(team, start, end), [=](const int idx, int &lsum) { lsum += (glb_Aai[idx + 1] - glb_Aai[idx]); }, nnz);
+    parallel_reduce(Kokkos::TeamVectorRange(team, start, end), [=](const int idx, int &lsum) { lsum += (glb_Aai[idx + 1] - glb_Aai[idx]); }, nnz);
     metad->flops = 2 * (metad->its * (10 * Nblk + 2 * nnz) + 5 * Nblk);
   } else {
     metad->flops = 2 * (metad->its * (10 * Nblk + 2 * 50 * Nblk) + 5 * Nblk); // guess
@@ -398,12 +391,11 @@ static KOKKOS_INLINE_FUNCTION PetscErrorCode BJSolve_BICG(const team_member team
   });
   team.team_barrier();
   /*    dp <- r'*r       */
-  parallel_reduce(
-    Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &lsum) { lsum += Rr[idx] * PetscConj(Rr[idx]); }, dpi);
+  parallel_reduce(Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &lsum) { lsum += Rr[idx] * PetscConj(Rr[idx]); }, dpi);
   team.team_barrier();
   r0 = dp = PetscSqrtReal(PetscRealPart(dpi));
 #if defined(PETSC_USE_DEBUG) && !defined(PETSC_HAVE_SYCL)
-  if (monitor) Kokkos::single(Kokkos::PerTeam(team), [=]() { printf("%3d KSP Residual norm %14.12e \n", 0, (double)dp); });
+  if (monitor) Kokkos::single(Kokkos::PerTeam(team), [=]() { printf("%3d KSP Residual norm %14.12e\n", 0, (double)dp); });
 #endif
   if (dp < atol) {
     metad->reason = KSP_CONVERGED_ATOL_NORMAL;
@@ -419,8 +411,7 @@ static KOKKOS_INLINE_FUNCTION PetscErrorCode BJSolve_BICG(const team_member team
   it = 0;
   do {
     /*     beta <- r'z     */
-    parallel_reduce(
-      Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &dot) { dot += Zr[idx] * PetscConj(Rl[idx]); }, beta);
+    parallel_reduce(Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &dot) { dot += Zr[idx] * PetscConj(Rl[idx]); }, beta);
     team.team_barrier();
 #if PCBJKOKKOS_VERBOSE_LEVEL >= 6
   #if defined(PETSC_USE_DEBUG) && !defined(PETSC_HAVE_SYCL)
@@ -452,8 +443,7 @@ static KOKKOS_INLINE_FUNCTION PetscErrorCode BJSolve_BICG(const team_member team
     static_cast<void>(MatMult(team, glb_Aai, glb_Aaj, glb_Aaa, r, ic, start, end, Pr, Zr));
     static_cast<void>(MatMultTranspose(team, glb_Aai, glb_Aaj, glb_Aaa, r, ic, start, end, Pl, Zl));
     /*     dpi <- z'p      */
-    parallel_reduce(
-      Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &lsum) { lsum += Zr[idx] * PetscConj(Pl[idx]); }, dpi);
+    parallel_reduce(Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &lsum) { lsum += Zr[idx] * PetscConj(Pl[idx]); }, dpi);
     team.team_barrier();
     if (dpi == 0) {
       metad->reason = KSP_CONVERGED_HAPPY_BREAKDOWN;
@@ -472,12 +462,11 @@ static KOKKOS_INLINE_FUNCTION PetscErrorCode BJSolve_BICG(const team_member team
     team.team_barrier();
     team.team_barrier();
     /*    dp <- r'*r       */
-    parallel_reduce(
-      Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &lsum) { lsum += Rr[idx] * PetscConj(Rr[idx]); }, dpi);
+    parallel_reduce(Kokkos::TeamVectorRange(team, Nblk), [=](const int idx, PetscScalar &lsum) { lsum += Rr[idx] * PetscConj(Rr[idx]); }, dpi);
     team.team_barrier();
     dp = PetscSqrtReal(PetscRealPart(dpi));
 #if defined(PETSC_USE_DEBUG) && !defined(PETSC_HAVE_SYCL)
-    if (monitor) Kokkos::single(Kokkos::PerTeam(team), [=]() { printf("%3d KSP Residual norm %14.12e \n", it + 1, (double)dp); });
+    if (monitor) Kokkos::single(Kokkos::PerTeam(team), [=]() { printf("%3d KSP Residual norm %14.12e\n", it + 1, (double)dp); });
 #endif
     if (dp < atol) {
       metad->reason = KSP_CONVERGED_ATOL_NORMAL;
@@ -523,8 +512,7 @@ done:
   metad->its = it;
   if (1) {
     int nnz;
-    parallel_reduce(
-      Kokkos::TeamVectorRange(team, start, end), [=](const int idx, int &lsum) { lsum += (glb_Aai[idx + 1] - glb_Aai[idx]); }, nnz);
+    parallel_reduce(Kokkos::TeamVectorRange(team, start, end), [=](const int idx, int &lsum) { lsum += (glb_Aai[idx + 1] - glb_Aai[idx]); }, nnz);
     metad->flops = 2 * (metad->its * (10 * Nblk + 2 * nnz) + 5 * Nblk);
   } else {
     metad->flops = 2 * (metad->its * (10 * Nblk + 2 * 50 * Nblk) + 5 * Nblk); // guess
@@ -585,7 +573,7 @@ static PetscErrorCode PCApply_BJKOKKOS(PC pc, Vec bin, Vec xout)
     // get x
     PetscCall(VecGetArrayAndMemType(xout, &glb_xdata, &mtype));
 #if defined(PETSC_HAVE_CUDA)
-    PetscCheck(PetscMemTypeDevice(mtype), PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "No GPU data for x %" PetscInt_FMT " != %" PetscInt_FMT, mtype, PETSC_MEMTYPE_DEVICE);
+    PetscCheck(PetscMemTypeDevice(mtype), PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "No GPU data for x %d != %d", static_cast<int>(mtype), static_cast<int>(PETSC_MEMTYPE_DEVICE));
 #endif
     PetscCall(VecGetArrayReadAndMemType(bvec, &glb_bdata, &mtype));
 #if defined(PETSC_HAVE_CUDA)
@@ -604,7 +592,7 @@ static PetscErrorCode PCApply_BJKOKKOS(PC pc, Vec bin, Vec xout)
 #if defined(PETSC_HAVE_KOKKOS_KERNELS_BATCH)
       PetscCall(PCApply_BJKOKKOSKERNELS(pc, glb_bdata, glb_xdata, glb_Aai, glb_Aaj, glb_Aaa, team_size, info, batch_sz, &pcreason));
 #else
-      PetscCheck(ksp_type_idx != BATCH_KSP_GMRESKK_IDX, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "Type: BATCH_KSP_GMRES not supported for complex\n");
+      PetscCheck(ksp_type_idx != BATCH_KSP_GMRESKK_IDX, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "Type: BATCH_KSP_GMRES not supported for complex");
 #endif
     } else { // Kokkos Krylov
       using scr_mem_t    = Kokkos::DefaultExecutionSpace::scratch_memory_space;
@@ -614,7 +602,7 @@ static PetscErrorCode PCApply_BJKOKKOS(PC pc, Vec bin, Vec xout)
       d_bid_eqOffset = jac->d_bid_eqOffset_k->data();
       // solve each block independently
       int scr_bytes_team_shared = 0, nShareVec = 0, nGlobBVec = 0;
-      if (jac->const_block_size) { // use shared memory for work vectors only if constant block size - todo: test efficiency loss
+      if (jac->const_block_size) { // use shared memory for work vectors only if constant block size - TODO: test efficiency loss
         size_t      maximum_shared_mem_size = 64000;
         PetscDevice device;
         PetscCall(PetscDeviceGetDefault_Internal(&device));
@@ -669,32 +657,43 @@ static PetscErrorCode PCApply_BJKOKKOS(PC pc, Vec bin, Vec xout)
 #endif
       auto h_metadata = Kokkos::create_mirror(Kokkos::HostSpace::memory_space(), d_metadata);
       Kokkos::deep_copy(h_metadata, d_metadata);
-      PetscInt count = -1, mbid = 0;
-      int      in[2], out[2];
+      PetscInt max_nnit = -1;
+#if PCBJKOKKOS_VERBOSE_LEVEL > 1
+      PetscInt mbid = 0;
+#endif
+      int in[2], out[2];
       if (jac->reason) { // -pc_bjkokkos_ksp_converged_reason
 #if PCBJKOKKOS_VERBOSE_LEVEL >= 3
   #if PCBJKOKKOS_VERBOSE_LEVEL >= 4
         PetscCall(PetscPrintf(PETSC_COMM_WORLD, "Iterations\n"));
   #endif
         // assume species major
-  #if PCBJKOKKOS_VERBOSE_LEVEL < 4
+  #if PCBJKOKKOS_VERBOSE_LEVEL == 3
         if (batch_sz != 1) PetscCall(PetscPrintf(PetscObjectComm((PetscObject)A), "%s: max iterations per species:", ksp_type_idx == BATCH_KSP_BICG_IDX ? "bicg" : "tfqmr"));
         else PetscCall(PetscPrintf(PetscObjectComm((PetscObject)A), "    Linear solve converged due to %s iterations ", ksp_type_idx == BATCH_KSP_BICG_IDX ? "bicg" : "tfqmr"));
   #endif
-        for (PetscInt dmIdx = 0, head = 0; dmIdx < jac->num_dms; dmIdx += batch_sz) {
-          for (PetscInt f = 0, idx = head; f < jac->dm_Nf[dmIdx]; f++, s++, idx++) {
+        for (PetscInt dmIdx = 0, head = 0, s = 0; dmIdx < jac->num_dms; dmIdx += batch_sz) {
+          for (PetscInt f = 0, idx = head; f < jac->dm_Nf[dmIdx]; f++, idx++, s++) {
+            for (int bid = 0; bid < batch_sz; bid++) {
+  #if PCBJKOKKOS_VERBOSE_LEVEL >= 4
+              jac->max_nits += h_metadata[idx + bid * jac->dm_Nf[dmIdx]].its; // report total number of iterations with high verbose
+              if (h_metadata[idx + bid * jac->dm_Nf[dmIdx]].its > max_nnit) {
+                max_nnit = h_metadata[idx + bid * jac->dm_Nf[dmIdx]].its;
+                mbid     = bid;
+              }
+  #else
+              if (h_metadata[idx + bid * jac->dm_Nf[dmIdx]].its > max_nnit) {
+                jac->max_nits = max_nnit = h_metadata[idx + bid * jac->dm_Nf[dmIdx]].its;
+                mbid                     = bid;
+              }
+  #endif
+            }
   #if PCBJKOKKOS_VERBOSE_LEVEL >= 4
             PetscCall(PetscPrintf(PetscObjectComm((PetscObject)A), "%2" PetscInt_FMT ":", s));
             for (int bid = 0; bid < batch_sz; bid++) PetscCall(PetscPrintf(PetscObjectComm((PetscObject)A), "%3" PetscInt_FMT " ", h_metadata[idx + bid * jac->dm_Nf[dmIdx]].its));
             PetscCall(PetscPrintf(PetscObjectComm((PetscObject)A), "\n"));
-  #else
-            for (int bid = 0; bid < batch_sz; bid++) {
-              if (h_metadata[idx + bid * jac->dm_Nf[dmIdx]].its > count) {
-                count = h_metadata[idx + bid * jac->dm_Nf[dmIdx]].its;
-                mbid  = bid;
-              }
-            }
-            PetscCall(PetscPrintf(PetscObjectComm((PetscObject)A), "%3" PetscInt_FMT " ", count));
+  #else // == 3
+            PetscCall(PetscPrintf(PetscObjectComm((PetscObject)A), "%3" PetscInt_FMT " ", max_nnit));
   #endif
           }
           head += batch_sz * jac->dm_Nf[dmIdx];
@@ -703,36 +702,31 @@ static PetscErrorCode PCApply_BJKOKKOS(PC pc, Vec bin, Vec xout)
         PetscCall(PetscPrintf(PetscObjectComm((PetscObject)A), "\n"));
   #endif
 #endif
-        if (count == -1) {
+        if (max_nnit == -1) { // < 3
           for (int blkID = 0; blkID < nBlk; blkID++) {
-            if (h_metadata[blkID].its > count) {
-              jac->max_nits = count = h_metadata[blkID].its;
-              mbid                  = blkID;
-            }
-#if PCBJKOKKOS_VERBOSE_LEVEL > 0
-            if (h_metadata[blkID].reason < 0) {
-              PetscCall(PetscPrintf(PETSC_COMM_SELF, "ERROR reason=%s, its=%" PetscInt_FMT ". species %" PetscInt_FMT ", batch %" PetscInt_FMT "\n", KSPConvergedReasons[h_metadata[blkID].reason], h_metadata[blkID].its, blkID / batch_sz, blkID % batch_sz));
-            }
+            if (h_metadata[blkID].its > max_nnit) {
+              jac->max_nits = max_nnit = h_metadata[blkID].its;
+#if PCBJKOKKOS_VERBOSE_LEVEL > 1
+              mbid = blkID;
 #endif
-            PetscCall(PetscLogGpuFlops((PetscLogDouble)h_metadata[blkID].flops));
+            }
           }
         }
-        in[0] = count;
+        in[0] = max_nnit;
         in[1] = rank;
-        PetscCallMPI(MPI_Allreduce(in, out, 1, MPI_2INT, MPI_MAXLOC, PetscObjectComm((PetscObject)A)));
+        PetscCallMPI(MPIU_Allreduce(in, out, 1, MPI_2INT, MPI_MAXLOC, PetscObjectComm((PetscObject)A)));
+#if PCBJKOKKOS_VERBOSE_LEVEL > 1
         if (0 == rank) {
           if (batch_sz != 1)
-            PetscCall(PetscPrintf(PETSC_COMM_SELF, "[%d] Linear solve converged due to %s iterations %d, batch %" PetscInt_FMT ", species %" PetscInt_FMT " (max)\n", out[1], KSPConvergedReasons[h_metadata[mbid].reason], out[0], mbid % batch_sz, mbid / batch_sz));
-          else PetscCall(PetscPrintf(PETSC_COMM_SELF, "[%d] Linear solve converged due to %s iterations %d, block %" PetscInt_FMT " (max)\n", out[1], KSPConvergedReasons[h_metadata[mbid].reason], out[0], mbid));
+            PetscCall(PetscPrintf(PETSC_COMM_SELF, "[%d] Linear solve converged due to %s iterations %d (max), on block %" PetscInt_FMT ", species %" PetscInt_FMT " (max)\n", out[1], KSPConvergedReasons[h_metadata[mbid].reason], out[0], mbid % batch_sz, mbid / batch_sz));
+          else PetscCall(PetscPrintf(PETSC_COMM_SELF, "[%d] Linear solve converged due to %s iterations %d (max), on block %" PetscInt_FMT "\n", out[1], KSPConvergedReasons[h_metadata[mbid].reason], out[0], mbid));
         }
+#endif
       }
       for (int blkID = 0; blkID < nBlk; blkID++) {
         PetscCall(PetscLogGpuFlops((PetscLogDouble)h_metadata[blkID].flops));
-#if PCBJKOKKOS_VERBOSE_LEVEL > 0
-        if (h_metadata[blkID].reason < 0) {
-          PetscCall(PetscPrintf(PETSC_COMM_SELF, "ERROR reason=%s, its=%" PetscInt_FMT ". species %" PetscInt_FMT ", batch %" PetscInt_FMT "\n", KSPConvergedReasons[h_metadata[blkID].reason], h_metadata[blkID].its, blkID / batch_sz, blkID % batch_sz));
-        }
-#endif
+        PetscCheck(h_metadata[blkID].reason >= 0 || !jac->ksp->errorifnotconverged, PetscObjectComm((PetscObject)pc), PETSC_ERR_CONV_FAILED, "ERROR reason=%s, its=%" PetscInt_FMT ". species %" PetscInt_FMT ", batch %" PetscInt_FMT,
+                   KSPConvergedReasons[h_metadata[blkID].reason], h_metadata[blkID].its, blkID / batch_sz, blkID % batch_sz);
       }
       {
         int errsum;
@@ -776,8 +770,7 @@ static PetscErrorCode PCSetUp_BJKOKKOS(PC pc)
   PetscBool      flg;
 
   PetscFunctionBegin;
-  //PetscCheck(!pc->useAmat, PetscObjectComm((PetscObject)pc), PETSC_ERR_SUP, "No support for using 'use_amat'");
-  PetscCheck(A, PetscObjectComm((PetscObject)A), PETSC_ERR_USER, "No matrix - A is used above");
+  PetscCheck(A, PetscObjectComm((PetscObject)A), PETSC_ERR_ARG_WRONG, "No matrix - A is used above");
   PetscCall(PetscObjectTypeCompareAny((PetscObject)A, &flg, MATSEQAIJKOKKOS, MATMPIAIJKOKKOS, MATAIJKOKKOS, ""));
   PetscCheck(flg, PetscObjectComm((PetscObject)A), PETSC_ERR_ARG_WRONG, "must use '-[dm_]mat_type aijkokkos -[dm_]vec_type kokkos' for -pc_type bjkokkos");
   if (!A->spptr) {
@@ -832,7 +825,7 @@ static PetscErrorCode PCSetUp_BJKOKKOS(PC pc)
           PetscCall(DMCreateGlobalVector(pack, &jac->vec_diag));
         } else pack = NULL; // flag for no DM
       }
-      if (!jac->vec_diag) { // get 'nDMs' and sizes 'block_sizes' w/o DMComposite. User could provide ISs (todo)
+      if (!jac->vec_diag) { // get 'nDMs' and sizes 'block_sizes' w/o DMComposite. TODO: User could provide ISs
         PetscInt        bsrt, bend, ncols, ntot = 0;
         const PetscInt *colsA, nloc = Iend - Istart;
         const PetscInt *rowindices, *icolindices;
@@ -843,14 +836,14 @@ static PetscErrorCode PCSetUp_BJKOKKOS(PC pc)
         bsrt = 0;
         bend = 1;
         for (PetscInt row_B = 0; row_B < nloc; row_B++) { // for all rows in block diagonal space
-          PetscInt rowA = icolindices[row_B], minj = PETSC_MAX_INT, maxj = 0;
+          PetscInt rowA = icolindices[row_B], minj = PETSC_INT_MAX, maxj = 0;
           //PetscCall(PetscPrintf(PETSC_COMM_SELF, "\t[%d] rowA = %d\n",rank,rowA));
           PetscCall(MatGetRow(Aseq, rowA, &ncols, &colsA, NULL)); // not sorted in permutation
-          PetscCheck(ncols, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "Empty row not supported: %" PetscInt_FMT "\n", row_B);
+          PetscCheck(ncols, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "Empty row not supported: %" PetscInt_FMT, row_B);
           for (PetscInt colj = 0; colj < ncols; colj++) {
             PetscInt colB = rowindices[colsA[colj]]; // use local idx
             //PetscCall(PetscPrintf(PETSC_COMM_SELF, "\t\t[%d] colB = %d\n",rank,colB));
-            PetscCheck(colB >= 0 && colB < nloc, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "colB < 0: %" PetscInt_FMT "\n", colB);
+            PetscCheck(colB >= 0 && colB < nloc, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "colB < 0: %" PetscInt_FMT, colB);
             if (colB > maxj) maxj = colB;
             if (colB < minj) minj = colB;
           }
@@ -859,13 +852,13 @@ static PetscErrorCode PCSetUp_BJKOKKOS(PC pc)
             //PetscCall(PetscPrintf(PetscObjectComm((PetscObject)A), "\t\t finish block %d, N loc = %d (%d,%d)\n", nDMs+1, bend - bsrt,bsrt,bend));
             block_sizes[nDMs] = bend - bsrt;
             ntot += block_sizes[nDMs];
-            PetscCheck(minj == bend, PETSC_COMM_SELF, PETSC_ERR_ARG_WRONG, "minj != bend: %" PetscInt_FMT " != %" PetscInt_FMT "\n", minj, bend);
+            PetscCheck(minj == bend, PETSC_COMM_SELF, PETSC_ERR_ARG_WRONG, "minj != bend: %" PetscInt_FMT " != %" PetscInt_FMT, minj, bend);
             bsrt = bend;
             bend++; // start with size 1 in new block
             nDMs++;
           }
           if (maxj + 1 > bend) bend = maxj + 1;
-          PetscCheck(minj >= bsrt || row_B == Iend - 1, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "%" PetscInt_FMT ") minj < bsrt: %" PetscInt_FMT " != %" PetscInt_FMT "\n", rowA, minj, bsrt);
+          PetscCheck(minj >= bsrt || row_B == Iend - 1, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "%" PetscInt_FMT ") minj < bsrt: %" PetscInt_FMT " != %" PetscInt_FMT, rowA, minj, bsrt);
           //PetscCall(PetscPrintf(PETSC_COMM_SELF, "[%d] %d) row %d.%d) cols %d : %d ; bsrt = %d, bend = %d\n",rank,row_B,nDMs,rowA,minj,maxj,bsrt,bend));
         }
         // do last block
@@ -874,7 +867,7 @@ static PetscErrorCode PCSetUp_BJKOKKOS(PC pc)
         ntot += block_sizes[nDMs];
         nDMs++;
         // cleanup
-        PetscCheck(ntot == nloc, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "n total != n local: %" PetscInt_FMT " != %" PetscInt_FMT "\n", ntot, nloc);
+        PetscCheck(ntot == nloc, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "n total != n local: %" PetscInt_FMT " != %" PetscInt_FMT, ntot, nloc);
         PetscCall(ISRestoreIndices(isrow, &rowindices));
         PetscCall(ISRestoreIndices(isicol, &icolindices));
         PetscCall(PetscRealloc(sizeof(PetscInt) * nDMs, &block_sizes));
@@ -908,7 +901,7 @@ static PetscErrorCode PCSetUp_BJKOKKOS(PC pc)
 #else
           KSPType ksptype;
           PetscCall(KSPGetType(jac->ksp, &ksptype));
-          PetscCheck(flg, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "Type: %s not supported in complex\n", ksptype);
+          PetscCheck(flg, PetscObjectComm((PetscObject)pc), PETSC_ERR_ARG_WRONG, "Type: %s not supported in complex", ksptype);
 #endif
         }
       }
@@ -1044,7 +1037,6 @@ static PetscErrorCode PCReset_BJKOKKOS(PC pc)
   jac->batch_b      = NULL;
   jac->batch_x      = NULL;
   jac->batch_values = NULL;
-
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
@@ -1091,7 +1083,7 @@ static PetscErrorCode PCBJKOKKOSSetKSP_BJKOKKOS(PC pc, KSP ksp)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-/*@C
+/*@
   PCBJKOKKOSSetKSP - Sets the `KSP` context for `PCBJKOKKOS`
 
   Collective
@@ -1107,7 +1099,7 @@ static PetscErrorCode PCBJKOKKOSSetKSP_BJKOKKOS(PC pc, KSP ksp)
 
   If the `PC` is not `PCBJKOKKOS` this function returns without doing anything
 
-  .seealso: `PCBJKOKKOSGetKSP()`, `PCBJKOKKOS`
+.seealso: [](ch_ksp), `PCBJKOKKOSGetKSP()`, `PCBJKOKKOS`
 @*/
 PetscErrorCode PCBJKOKKOSSetKSP(PC pc, KSP ksp)
 {
@@ -1129,7 +1121,7 @@ static PetscErrorCode PCBJKOKKOSGetKSP_BJKOKKOS(PC pc, KSP *ksp)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-/*@C
+/*@
   PCBJKOKKOSGetKSP - Gets the `KSP` context for the `PCBJKOKKOS` preconditioner
 
   Not Collective but `KSP` returned is parallel if `PC` was parallel
@@ -1147,7 +1139,7 @@ static PetscErrorCode PCBJKOKKOSGetKSP_BJKOKKOS(PC pc, KSP *ksp)
 
   If the `PC` is not a `PCBJKOKKOS` object it raises an error
 
-.seealso: `PCBJKOKKOS`, `PCBJKOKKOSSetKSP()`
+.seealso: [](ch_ksp), `PCBJKOKKOS`, `PCBJKOKKOSSetKSP()`
 @*/
 PetscErrorCode PCBJKOKKOSGetKSP(PC pc, KSP *ksp)
 {
@@ -1158,6 +1150,26 @@ PetscErrorCode PCBJKOKKOSGetKSP(PC pc, KSP *ksp)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
+static PetscErrorCode PCPostSolve_BJKOKKOS(PC pc, KSP ksp, Vec b, Vec x)
+{
+  PC_PCBJKOKKOS *jac = (PC_PCBJKOKKOS *)pc->data;
+
+  PetscFunctionBegin;
+  PetscValidHeaderSpecific(pc, PC_CLASSID, 1);
+  ksp->its = jac->max_nits;
+  PetscFunctionReturn(PETSC_SUCCESS);
+}
+
+static PetscErrorCode PCPreSolve_BJKOKKOS(PC pc, KSP ksp, Vec b, Vec x)
+{
+  PC_PCBJKOKKOS *jac = (PC_PCBJKOKKOS *)pc->data;
+
+  PetscFunctionBegin;
+  PetscValidHeaderSpecific(pc, PC_CLASSID, 1);
+  jac->ksp->errorifnotconverged = ksp->errorifnotconverged;
+  PetscFunctionReturn(PETSC_SUCCESS);
+}
+
 /*MC
      PCBJKOKKOS -  Defines a preconditioner that applies a Krylov solver and preconditioner to the blocks in a `MATSEQAIJ` matrix on the GPU using Kokkos
 
@@ -1174,7 +1186,7 @@ PetscErrorCode PCBJKOKKOSGetKSP(PC pc, KSP *ksp)
 
    Why does it have its own `KSP`? Where is the `KSP` run if used with -ksp_type preonly?
 
-.seealso: `PCCreate()`, `PCSetType()`, `PCType`, `PC`, `PCBJACOBI`,
+.seealso: [](ch_ksp), `PCCreate()`, `PCSetType()`, `PCType`, `PC`, `PCBJACOBI`,
           `PCSHELL`, `PCCOMPOSITE`, `PCSetUseAmat()`, `PCBJKOKKOSGetKSP()`
 M*/
 
@@ -1203,6 +1215,8 @@ PETSC_EXTERN PetscErrorCode PCCreate_BJKOKKOS(PC pc)
   pc->ops->destroy        = PCDestroy_BJKOKKOS;
   pc->ops->setfromoptions = PCSetFromOptions_BJKOKKOS;
   pc->ops->view           = PCView_BJKOKKOS;
+  pc->ops->postsolve      = PCPostSolve_BJKOKKOS;
+  pc->ops->presolve       = PCPreSolve_BJKOKKOS;
 
   jac->rowOffsets   = NULL;
   jac->colIndices   = NULL;
