@@ -51,6 +51,7 @@ PetscErrorCode VecDuplicate_MPI(Vec win, Vec *v)
   PetscCall(PetscObjectListDuplicate(((PetscObject)win)->olist, &((PetscObject)*v)->olist));
   PetscCall(PetscFunctionListDuplicate(((PetscObject)win)->qlist, &((PetscObject)*v)->qlist));
 
+  (*v)->stash.bs  = win->stash.bs;
   (*v)->bstash.bs = win->bstash.bs;
   PetscFunctionReturn(PETSC_SUCCESS);
 }
@@ -62,29 +63,28 @@ static PetscErrorCode VecDuplicateVecs_MPI_GEMV(Vec w, PetscInt m, Vec *V[])
   PetscFunctionBegin;
   // Currently only do GEMV for vectors without ghosts. Note w might be a VECMPI subclass object.
   // This routine relies on the duplicate operation being VecDuplicate_MPI. If not, bail out to the default.
-  if (wmpi->nghost || w->ops->duplicate != VecDuplicate_MPI) {
+  if (wmpi->localrep || w->ops->duplicate != VecDuplicate_MPI) {
     w->ops->duplicatevecs = VecDuplicateVecs_Default;
     PetscCall(VecDuplicateVecs(w, m, V));
   } else {
-    PetscInt     nlocal;
     PetscScalar *array;
     PetscInt64   lda; // use 64-bit as we will do "m * lda"
 
     PetscCall(PetscMalloc1(m, V));
-    PetscCall(VecGetLocalSize(w, &nlocal));
-    lda = nlocal;
-    lda = ((lda + 31) / 32) * 32; // make every vector 32-elements aligned
+    VecGetLocalSizeAligned(w, 64, &lda); // get in lda the 64-bytes aligned local size
 
     PetscCall(PetscCalloc1(m * lda, &array));
     for (PetscInt i = 0; i < m; i++) {
       Vec v;
       PetscCall(VecCreateMPIWithLayoutAndArray_Private(w->map, PetscSafePointerPlusOffset(array, i * lda), &v));
+      PetscCall(VecSetType(v, ((PetscObject)w)->type_name));
       PetscCall(PetscObjectListDuplicate(((PetscObject)w)->olist, &((PetscObject)v)->olist));
       PetscCall(PetscFunctionListDuplicate(((PetscObject)w)->qlist, &((PetscObject)v)->qlist));
-      v->ops->view          = w->ops->view;
+      v->ops[0]             = w->ops[0];
       v->stash.donotstash   = w->stash.donotstash;
       v->stash.ignorenegidx = w->stash.ignorenegidx;
       v->stash.bs           = w->stash.bs;
+      v->bstash.bs          = w->bstash.bs;
       (*V)[i]               = v;
     }
     // So when the first vector is destroyed it will destroy the array
@@ -143,12 +143,12 @@ static PetscErrorCode VecAssemblySend_MPI_Private(MPI_Comm comm, const PetscMPII
      receiver is expecting them.
    */
   if (hdr->count || (x->first_assembly_done && x->sendptrs[rankid].ints)) {
-    PetscCallMPI(MPI_Isend(x->sendptrs[rankid].ints, hdr->count, MPIU_INT, rank, tag[0], comm, &req[0]));
-    PetscCallMPI(MPI_Isend(x->sendptrs[rankid].scalars, hdr->count, MPIU_SCALAR, rank, tag[1], comm, &req[1]));
+    PetscCallMPI(MPIU_Isend(x->sendptrs[rankid].ints, hdr->count, MPIU_INT, rank, tag[0], comm, &req[0]));
+    PetscCallMPI(MPIU_Isend(x->sendptrs[rankid].scalars, hdr->count, MPIU_SCALAR, rank, tag[1], comm, &req[1]));
   }
   if (hdr->bcount || (x->first_assembly_done && x->sendptrs[rankid].intb)) {
-    PetscCallMPI(MPI_Isend(x->sendptrs[rankid].intb, hdr->bcount, MPIU_INT, rank, tag[2], comm, &req[2]));
-    PetscCallMPI(MPI_Isend(x->sendptrs[rankid].scalarb, hdr->bcount * bs, MPIU_SCALAR, rank, tag[3], comm, &req[3]));
+    PetscCallMPI(MPIU_Isend(x->sendptrs[rankid].intb, hdr->bcount, MPIU_INT, rank, tag[2], comm, &req[2]));
+    PetscCallMPI(MPIU_Isend(x->sendptrs[rankid].scalarb, hdr->bcount * bs, MPIU_SCALAR, rank, tag[3], comm, &req[3]));
   }
   PetscFunctionReturn(PETSC_SUCCESS);
 }
@@ -166,9 +166,9 @@ static PetscErrorCode VecAssemblyRecv_MPI_Private(MPI_Comm comm, const PetscMPII
 
   if (hdr->count) {
     PetscCall(PetscSegBufferGet(x->segrecvint, hdr->count, &frame->ints));
-    PetscCallMPI(MPI_Irecv(frame->ints, hdr->count, MPIU_INT, rank, tag[0], comm, &req[0]));
+    PetscCallMPI(MPIU_Irecv(frame->ints, hdr->count, MPIU_INT, rank, tag[0], comm, &req[0]));
     PetscCall(PetscSegBufferGet(x->segrecvscalar, hdr->count, &frame->scalars));
-    PetscCallMPI(MPI_Irecv(frame->scalars, hdr->count, MPIU_SCALAR, rank, tag[1], comm, &req[1]));
+    PetscCallMPI(MPIU_Irecv(frame->scalars, hdr->count, MPIU_SCALAR, rank, tag[1], comm, &req[1]));
     frame->pendings = 2;
   } else {
     frame->ints     = NULL;
@@ -178,9 +178,9 @@ static PetscErrorCode VecAssemblyRecv_MPI_Private(MPI_Comm comm, const PetscMPII
 
   if (hdr->bcount) {
     PetscCall(PetscSegBufferGet(x->segrecvint, hdr->bcount, &frame->intb));
-    PetscCallMPI(MPI_Irecv(frame->intb, hdr->bcount, MPIU_INT, rank, tag[2], comm, &req[2]));
+    PetscCallMPI(MPIU_Irecv(frame->intb, hdr->bcount, MPIU_INT, rank, tag[2], comm, &req[2]));
     PetscCall(PetscSegBufferGet(x->segrecvscalar, hdr->bcount * bs, &frame->scalarb));
-    PetscCallMPI(MPI_Irecv(frame->scalarb, hdr->bcount * bs, MPIU_SCALAR, rank, tag[3], comm, &req[3]));
+    PetscCallMPI(MPIU_Irecv(frame->scalarb, hdr->bcount * bs, MPIU_SCALAR, rank, tag[3], comm, &req[3]));
     frame->pendingb = 2;
   } else {
     frame->intb     = NULL;
@@ -192,9 +192,10 @@ static PetscErrorCode VecAssemblyRecv_MPI_Private(MPI_Comm comm, const PetscMPII
 
 static PetscErrorCode VecAssemblyBegin_MPI_BTS(Vec X)
 {
-  Vec_MPI *x = (Vec_MPI *)X->data;
-  MPI_Comm comm;
-  PetscInt i, j, jb, bs;
+  Vec_MPI    *x = (Vec_MPI *)X->data;
+  MPI_Comm    comm;
+  PetscMPIInt i;
+  PetscInt    j, jb, bs;
 
   PetscFunctionBegin;
   if (X->stash.donotstash) PetscFunctionReturn(PETSC_SUCCESS);
@@ -203,7 +204,8 @@ static PetscErrorCode VecAssemblyBegin_MPI_BTS(Vec X)
   PetscCall(VecGetBlockSize(X, &bs));
   if (PetscDefined(USE_DEBUG)) {
     InsertMode addv;
-    PetscCall(MPIU_Allreduce((PetscEnum *)&X->stash.insertmode, (PetscEnum *)&addv, 1, MPIU_ENUM, MPI_BOR, comm));
+
+    PetscCallMPI(MPIU_Allreduce((PetscEnum *)&X->stash.insertmode, (PetscEnum *)&addv, 1, MPIU_ENUM, MPI_BOR, comm));
     PetscCheck(addv != (ADD_VALUES | INSERT_VALUES), comm, PETSC_ERR_ARG_NOTSAMETYPE, "Some processors inserted values while others added");
   }
   X->bstash.insertmode = X->stash.insertmode; /* Block stash implicitly tracks InsertMode of scalar stash */
@@ -214,17 +216,19 @@ static PetscErrorCode VecAssemblyBegin_MPI_BTS(Vec X)
   if (!x->sendranks) {
     PetscMPIInt nowners, bnowners, *owners, *bowners;
     PetscInt    ntmp;
+
     PetscCall(VecStashGetOwnerList_Private(&X->stash, X->map, &nowners, &owners));
     PetscCall(VecStashGetOwnerList_Private(&X->bstash, X->map, &bnowners, &bowners));
     PetscCall(PetscMergeMPIIntArray(nowners, owners, bnowners, bowners, &ntmp, &x->sendranks));
-    x->nsendranks = ntmp;
+    PetscCall(PetscMPIIntCast(ntmp, &x->nsendranks));
     PetscCall(PetscFree(owners));
     PetscCall(PetscFree(bowners));
     PetscCall(PetscMalloc1(x->nsendranks, &x->sendhdr));
     PetscCall(PetscCalloc1(x->nsendranks, &x->sendptrs));
   }
   for (i = 0, j = 0, jb = 0; i < x->nsendranks; i++) {
-    PetscMPIInt rank         = x->sendranks[i];
+    PetscMPIInt rank = x->sendranks[i];
+
     x->sendhdr[i].insertmode = X->stash.insertmode;
     /* Initialize pointers for non-empty stashes the first time around.  Subsequent assemblies with
      * VEC_SUBSET_OFF_PROC_ENTRIES will leave the old pointers (dangling because the stash has been collected) when
@@ -288,7 +292,7 @@ static PetscErrorCode VecAssemblyEnd_MPI_BTS(Vec X)
     PetscFunctionReturn(PETSC_SUCCESS);
   }
 
-  PetscCheck(x->segrecvframe, PETSC_COMM_SELF, PETSC_ERR_PLIB, "Missing segrecvframe! Probably you forgot to call VecAssemblyBegin first");
+  PetscCheck(x->segrecvframe, PETSC_COMM_SELF, PETSC_ERR_PLIB, "Missing segrecvframe! Probably you forgot to call VecAssemblyBegin() first");
   PetscCall(VecGetArray(X, &xarray));
   PetscCall(PetscSegBufferExtractInPlace(x->segrecvframe, &frame));
   PetscCall(PetscMalloc2(4 * x->nrecvranks, &some_indices, x->use_status ? 4 * x->nrecvranks : 0, &some_statuses));
@@ -307,15 +311,20 @@ static PetscErrorCode VecAssemblyEnd_MPI_BTS(Vec X)
       PetscScalar *recvscalar;
       PetscBool    intmsg   = (PetscBool)(some_indices[ii] % 2 == 0);
       PetscBool    blockmsg = (PetscBool)((some_indices[ii] % 4) / 2 == 1);
+
       npending--;
       if (!blockmsg) { /* Scalar stash */
         PetscMPIInt count;
+
         if (--frame[i].pendings > 0) continue;
         if (x->use_status) {
           PetscCallMPI(MPI_Get_count(&some_statuses[ii], intmsg ? MPIU_INT : MPIU_SCALAR, &count));
-        } else count = x->recvhdr[i].count;
+        } else {
+          PetscCall(PetscMPIIntCast(x->recvhdr[i].count, &count));
+        }
         for (j = 0, recvint = frame[i].ints, recvscalar = frame[i].scalars; j < count; j++, recvint++) {
           PetscInt loc = *recvint - X->map->rstart;
+
           PetscCheck(*recvint >= X->map->rstart && X->map->rend > *recvint, PETSC_COMM_SELF, PETSC_ERR_PLIB, "Received vector entry %" PetscInt_FMT " out of local range [%" PetscInt_FMT ",%" PetscInt_FMT ")]", *recvint, X->map->rstart, X->map->rend);
           switch (imode) {
           case ADD_VALUES:
@@ -334,7 +343,9 @@ static PetscErrorCode VecAssemblyEnd_MPI_BTS(Vec X)
         if (x->use_status) {
           PetscCallMPI(MPI_Get_count(&some_statuses[ii], intmsg ? MPIU_INT : MPIU_SCALAR, &count));
           if (!intmsg) count /= bs; /* Convert from number of scalars to number of blocks */
-        } else count = x->recvhdr[i].bcount;
+        } else {
+          PetscCall(PetscMPIIntCast(x->recvhdr[i].bcount, &count));
+        }
         for (j = 0, recvint = frame[i].intb, recvscalar = frame[i].scalarb; j < count; j++, recvint++) {
           PetscInt loc = (*recvint) * bs - X->map->rstart;
           switch (imode) {
@@ -412,7 +423,6 @@ static PetscErrorCode VecGetLocalToGlobalMapping_MPI_VecGhost(Vec X, ISLocalToGl
   PetscInt       *indices, n, nghost, rstart, i;
   IS              ghostis;
   const PetscInt *ghostidx;
-  MPI_Comm        comm;
 
   PetscFunctionBegin;
   if (X->map->mapping) {
@@ -426,11 +436,10 @@ static PetscErrorCode VecGetLocalToGlobalMapping_MPI_VecGhost(Vec X, ISLocalToGl
   /* set local to global mapping for ghosted vector */
   PetscCall(PetscMalloc1(n + nghost, &indices));
   PetscCall(VecGetOwnershipRange(X, &rstart, NULL));
-  for (i = 0; i < n; i++) { indices[i] = rstart + i; }
-  for (i = 0; i < nghost; i++) { indices[n + i] = ghostidx[i]; }
+  for (i = 0; i < n; i++) indices[i] = rstart + i;
+  PetscCall(PetscArraycpy(indices + n, ghostidx, nghost));
   PetscCall(ISRestoreIndices(ghostis, &ghostidx));
-  PetscCall(PetscObjectGetComm((PetscObject)X, &comm));
-  PetscCall(ISLocalToGlobalMappingCreate(comm, 1, n + nghost, indices, PETSC_OWN_POINTER, &X->map->mapping));
+  PetscCall(ISLocalToGlobalMappingCreate(PETSC_COMM_SELF, 1, n + nghost, indices, PETSC_OWN_POINTER, &X->map->mapping));
   *ltg = X->map->mapping;
   PetscFunctionReturn(PETSC_SUCCESS);
 }
@@ -552,11 +561,10 @@ PetscErrorCode VecCreate_MPI_Private(Vec v, PetscBool alloc, PetscInt nghost, co
   if (mdot_use_gemv || maxpy_use_gemv) v->ops[0].duplicatevecs = VecDuplicateVecs_MPI_GEMV;
 
   if (mdot_use_gemv) {
-    v->ops[0].duplicatevecs = VecDuplicateVecs_MPI_GEMV;
-    v->ops[0].mdot          = VecMDot_MPI_GEMV;
-    v->ops[0].mdot_local    = VecMDot_Seq_GEMV;
-    v->ops[0].mtdot         = VecMTDot_MPI_GEMV;
-    v->ops[0].mtdot_local   = VecMTDot_Seq_GEMV;
+    v->ops[0].mdot        = VecMDot_MPI_GEMV;
+    v->ops[0].mdot_local  = VecMDot_Seq_GEMV;
+    v->ops[0].mtdot       = VecMTDot_MPI_GEMV;
+    v->ops[0].mtdot_local = VecMTDot_Seq_GEMV;
   }
   if (maxpy_use_gemv) v->ops[0].maxpy = VecMAXPY_Seq_GEMV;
 
@@ -661,7 +669,7 @@ PETSC_EXTERN PetscErrorCode VecCreate_Standard(Vec v)
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-/*@C
+/*@
   VecCreateMPIWithArray - Creates a parallel, array-style vector,
   where the user provides the array space to store the vector values.
 
@@ -705,7 +713,7 @@ PetscErrorCode VecCreateMPIWithArray(MPI_Comm comm, PetscInt bs, PetscInt n, Pet
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-/*@C
+/*@
   VecCreateGhostWithArray - Creates a parallel vector with ghost padding on each processor;
   the caller allocates the array space.
 
@@ -742,8 +750,6 @@ PetscErrorCode VecCreateGhostWithArray(MPI_Comm comm, PetscInt n, PetscInt N, Pe
 
   PetscFunctionBegin;
   *vv = NULL;
-
-  PetscCheck(n != PETSC_DECIDE, PETSC_COMM_SELF, PETSC_ERR_ARG_OUTOFRANGE, "Must set local size");
   PetscCheck(nghost != PETSC_DECIDE, PETSC_COMM_SELF, PETSC_ERR_ARG_OUTOFRANGE, "Must set local ghost size");
   PetscCheck(nghost >= 0, PETSC_COMM_SELF, PETSC_ERR_ARG_OUTOFRANGE, "Ghost length must be >= 0");
   PetscCall(PetscSplitOwnership(comm, &n, &N));
@@ -793,7 +799,7 @@ PetscErrorCode VecGhostGetGhostIS(Vec X, IS *ghost)
   PetscAssertPointer(ghost, 2);
   PetscCall(PetscObjectTypeCompare((PetscObject)X, VECMPI, &flg));
   PetscCheck(flg, PetscObjectComm((PetscObject)X), PETSC_ERR_ARG_WRONGSTATE, "VecGhostGetGhostIS was not supported for vec type %s", ((PetscObject)X)->type_name);
-  w      = (Vec_MPI *)(X)->data;
+  w      = (Vec_MPI *)X->data;
   *ghost = w->ghost;
   PetscFunctionReturn(PETSC_SUCCESS);
 }
@@ -879,7 +885,7 @@ PetscErrorCode VecMPISetGhost(Vec vv, PetscInt nghost, const PetscInt ghosts[])
     PetscUseTypeMethod(vv, destroy);
     PetscCall(VecSetSizes(vv, n, N));
     PetscCall(VecCreate_MPI_Private(vv, PETSC_TRUE, nghost, NULL));
-    w = (Vec_MPI *)(vv)->data;
+    w = (Vec_MPI *)vv->data;
     /* Create local representation */
     PetscCall(VecGetArray(vv, &larray));
     PetscCall(VecCreateSeqWithArray(PETSC_COMM_SELF, 1, n + nghost, larray, &w->localrep));
@@ -902,8 +908,7 @@ PetscErrorCode VecMPISetGhost(Vec vv, PetscInt nghost, const PetscInt ghosts[])
   PetscFunctionReturn(PETSC_SUCCESS);
 }
 
-/* ------------------------------------------------------------------------------------------*/
-/*@C
+/*@
   VecCreateGhostBlockWithArray - Creates a parallel vector with ghost padding on each processor;
   the caller allocates the array space. Indices in the ghost region are based on blocks.
 
